#include "StringBasics.h"
#include "StringArray.h"
#include "VariantCallFiles.h"
#include "Parameters.h"
#include "FunctionTimer.h"
#include <iostream>
#include "annotateLog.h"

int main (int argc, char ** const argv) {
	String inputVcfName("LDL.test.vcf");
	String inputLog(""); //LDL.test.log
	String output012("subsetGeno");
	String geneList(""); // LDL.geneList.txt
	String subsetID(""); // LDL.AA.txt
	int variantsChoice = 0;
	ParameterList pl;
	
	//command line specification
	// todo: 1, convert ped file into 012 matrix
	BEGIN_LONG_PARAMETERS(options)
		LONG_PARAMETER_GROUP("Input files")
		LONG_STRINGPARAMETER("vcf", &inputVcfName)
		LONG_STRINGPARAMETER("log",&inputLog)
		LONG_STRINGPARAMETER("IDfile", &subsetID)
		LONG_PARAMETER_GROUP("Subsetting choices")
		LONG_INTPARAMETER("All", &variantsChoice)	//	All:1, NonsynStop:0
		LONG_PARAMETER_GROUP("Output files")
		LONG_STRINGPARAMETER("outputPrefix",&output012)
		LONG_STRINGPARAMETER("outputGeneList",&geneList)
	END_LONG_PARAMETERS();
	
	pl.Add((new LongParameters("Data File PREPARATION",options))->SetPrecision(3));
	pl.Read(argc, argv);
	pl.Status();
	
	if (inputLog != "") {
		GeneMap testLog;
		printf("Read in the annotation information from file (%s)......\n",inputLog.c_str());
		printf("Write out the gene list into file (%s)......\n",geneList.c_str());
		AnnotateLog::Output_GeneInfo(inputLog,testLog,geneList);
		printf("Done.\n");		
	}
	
	// variantCallForm initalization
	VariantCallFormat test(inputVcfName);
	FunctionTimer recordTime;
	printf("Read in the vcf file (%s) ......\n", inputVcfName.c_str());
	printf("Write out the 012 genotype matrix files %s (.012, .012.pos, .012.indv, .012.frq) \n",output012.c_str());
	test.VCF2TransformedGenoNonsynStop(output012, false, subsetID, variantsChoice);
	printf("Done.\n");
	std::cout << "It takes " << recordTime.StopTime()/1e6 << " seconds" << std::endl;
	
	return 0;
	
	
/*	// output the size of the 012 matrix
	std::cout << test.nSubjects << std::endl; 
	test.VCFpositions();
	recordTime.Reset();
	std::cout << "total # of positions = " << test.numPos << std::endl;
	std::cout << "It takes " << recordTime.StopTime()/1e6 << " seconds" << std::endl;
	
	recordTime.Reset();
	test.VCFpositionsNonsynStop();
	std::cout << "total # of Nonsy and Stop positions = " << test.numPos << std::endl;	
	std::cout << "It takes " << recordTime.StopTime()/1e6 << " seconds" << std::endl;
	
	
	return 1;
	if (subsetID == "") { // if subsetID is not specified, then we do not need to subset the ID
		if (variantsChoice ==	0) { // if variantsChoice is 0, output all positions
			test.VCF2genotypes(output012);
		} else if (variantsChoice == 1){
			test.VCF2genotypesNonsynStop(output012);
		} else {
			printf("Invalid variantsChoice!\n");
			throw;
		}
	} else {
		if (variantsChoice == 0) {
			test.VCF2genotypes(output012,subsetID);
		} else if (variantsChoice == 1) {
			test.VCF2genotypesNonsynStop(output012,subsetID);
		} else {
			printf("Invalid variantsChoice!\n");
			throw;
		}
	}
	
	std::cout << "The time taken to conver the vcf file is " << recordTime.StopTime()/1e6 << " seconds"  << std::endl;
	return 1; */
}
