/*
 *  VariantCallFiles.cpp
 *  vcfReader
 *
 *  Created by Youna Hu on 1/17/11.
 *  Copyright 2011 U of Michigan. All rights reserved.
 *
 */

#include "VariantCallFiles.h"

void VariantCallFormat::VCFnumOutSubject(){
	VCFsubjects();
	nSubjects = subjectColumn.size();
}

void VariantCallFormat::VCFnumOutSubject(String &subjectsFile){
	DataFromFiles subjectIDdata;
	subjectIDdata.IDReadAsString(subjectsFile);
	nSubjects = subjectIDdata.GetIndvsIDs().size();
}

void VariantCallFormat::VCFpositions(){
	FILE *vcfIn = fopen(m_vcfName, "r");
	if (vcfIn == NULL) {
		 error("Error in opening file (%s)\n",m_vcfName.c_str());
	}
	numPos = 0;
	while (!feof(vcfIn)){
		fileEntry.ReadLine(vcfIn);
		if (fileEntry.IsEmpty()) {
			error("Error: File %s has an empty line (%d)\n",m_vcfName.c_str(),fileStep);
		}		
		if (fileEntry.Left(1) != "#") {
			numPos ++; // count the number of lines without # (the position lines)
		}
	}
}

void VariantCallFormat::VCFpositionsNonsynStop(){
	FILE *vcfIn = fopen(m_vcfName, "r");
	if (vcfIn == NULL) {
		error("Error in opening file (%s)\n",m_vcfName.c_str());
	}
	numPos = 0;
	while (!feof(vcfIn)){
		fileEntry.ReadLine(vcfIn);
		if (fileEntry.IsEmpty()) {
			error("Error: File %s has an empty line (%d)\n",m_vcfName.c_str(),fileStep);
		}		
		if (fileEntry.Left(1) != "#") { // count the number of lines without # (the position lines)
			fileToken.AddTokens(fileEntry, "\t");
			annotationInfo = fileToken[annotationCol].c_str();
			if (annotationInfo.find(searchNonsyn) != std::string::npos 
					|| annotationInfo.find(searchStop) != std::string::npos) {
				numPos ++; 
			}
		}
		fileEntry.Clear();
		fileToken.Clear();
	}
}

void VariantCallFormat::Init(String &vcfName){
	m_vcfName = vcfName;
	fileStep = 0;
	searchNonsyn = ";NON_SYN";
	searchStop = ";STOP";
	genoSum = 0;
	nSample = 0;
	genoStartCol = 9; // genotyep data starts at the 10th column
	idCol = 2; // ID is at the 3rd column
	refCol = 3; // reference allele is the 4th column
	altCol = 4; // alternative allele is the 5th column
	annotationCol = 7; // annotation is the 8th column 
	sampleByPos = false; // sampleByPos: rows -- samples, cols -- Positions
	VCFnumOutSubject();
	numPos = 0;
}

void VariantCallFormat::Init(String &vcfName, String &subjectsFile){
	m_vcfName = vcfName;
	fileStep = 0;
	searchNonsyn = ";NON_SYN";
	searchStop = ";STOP";
	genoSum = 0;
	nSample = 0;
	genoStartCol = 9; // genotyep data starts at the 10th column
	idCol = 2; // ID is at the 3rd column
	refCol = 3; // reference allele is the 4th column
	altCol = 4; // alternative allele is the 5th column
	annotationCol = 7; // annotation is the 8th column 
	sampleByPos = false; // sampleByPos: rows -- samples, cols -- Positions
	VCFnumOutSubject(subjectsFile);
	numPos = 0;
}

VariantCallFormat::VariantCallFormat(String &vcfName){
	Init(vcfName);
}

VariantCallFormat::VariantCallFormat(String &vcfName, String &subjectsFile){
	Init(vcfName,subjectsFile);
}

void VariantCallFormat::VCFsubjects(){
	FILE *vcfIn = fopen(m_vcfName, "r");
	if (vcfIn == NULL) {
		error("Error in opening file (%s)\n",m_vcfName.c_str());
	}
	fileStep = 0;
	while (!feof(vcfIn)) {
		fileEntry.ReadLine(vcfIn);
		if (fileEntry.IsEmpty()) {
			error("Error: File %s has an empty line (%d)\n",m_vcfName.c_str(),fileStep);
		}
		// find the line starts with $CHROM and read the subject IDs
		if (fileEntry.Left(6) == "#CHROM") {
		//	printf("Found %s\n",fileEntry.Left(6).c_str());
			fileToken.AddTokens(fileEntry, "\t");
			for (int i = genoStartCol; i < fileToken.Length(); i++) {
				subjectColumn.insert(std::map<String,int>::value_type(fileToken[i],i));
			}
			break;
		}
	}
	fclose(vcfIn);
	// printf("Number of subject %d\n",(int) subjectColumn.size());
}

bool VariantCallFormat::IsMissing(){
	if (alleleToken[0] != "0" && alleleToken[0] != "1" && alleleToken[1] != "0" && alleleToken[1] != "1") {
		return (true);
	} else {
		return (false);
	}
}

void VariantCallFormat::VCF2genotypes(String &genofileName){
	FILE *vcfIn = fopen(m_vcfName, "r");
	FILE *genoOut = fopen(genofileName + ".012","w");
	FILE *genoPos = fopen(genofileName + ".012.pos", "w");
	FILE *genoIndv = fopen(genofileName + ".012.indv", "w");
	FILE *genoAF = fopen(genofileName + ".012.frq", "w");
	
	if (vcfIn == NULL) {
		error("Error in opening file (%s)\n",m_vcfName.c_str());
	}
	
	if (genoOut == NULL){
		error("Error in opening file (%s)\n",genofileName.c_str());
	}
	
	fileStep = 0;
	
	while (!feof(vcfIn)) {
		fileEntry.ReadLine(vcfIn);
		if (fileEntry.IsEmpty()) {
			error("Error: File %s has an empty line (%d)\n",m_vcfName.c_str(),fileStep);
		}
		if (fileEntry.Left(6) == "#CHROM") {
			// write the indv file -- file with subject IDs
			fileToken.AddTokens(fileEntry, "\t");
			for (int i = 9; i < fileToken.Length(); i++) {
				fprintf(genoIndv, "%s\n", fileToken[i].c_str());
			}
			fclose(genoIndv);
		}
		if (fileEntry.Left(1) != "#") { // genotype contents start from the lines without # 
			fileToken.AddTokens(fileEntry, "\t");
			// write the position file
			fprintf(genoPos, "%s\t%s\n", fileToken[0].c_str(),fileToken[1].c_str());
			// write the genotypes and initialize the genoSum counts
			genoSum = 0;
			nSample = 0;
			for (int i = genoStartCol; i < fileToken.Length(); i++) {
				// get the genotype
				genoToken.AddTokens(fileToken[i], ":");
				//printf("genotype = %s\n",genoToken[0].c_str());
				alleleToken.AddTokens(genoToken[0], "/"); 
				
				// record the genotype -- check if it is misisng first
				if (!IsMissing()) {
					genoTypeNum = alleleToken[0].AsInteger() + alleleToken[1].AsInteger();	
					genoSum += genoTypeNum;
					nSample += 1;
				} else {
					genoTypeNum = -1;
				}
				
				// write out the genotype
				fprintf(genoOut, "%ld\t", genoTypeNum);
			
				// increment the file length
				fileStep = fileStep + 1;
				
				// clear the genoToken
				genoToken.Clear();
				alleleToken.Clear();
			}
			// write out the allele frequency (ID, ref, alt, af)
			double dSample = static_cast<double>(nSample);
			double dGenoSum = static_cast<double>(genoSum);
			fprintf(genoAF, "%s\t%s\t%s\t%1.5e\n", fileToken[idCol].c_str(), fileToken[refCol].c_str(), 
							fileToken[altCol].c_str(), 0.5*dGenoSum/dSample);
			fprintf(genoOut, "\n");
		}
		
		fileEntry.Clear();
		fileToken.Clear();
	}
	
	fclose(genoOut);
	fclose(genoPos);
	fclose(genoIndv);
	fclose(vcfIn);
	fclose(genoAF);
}	

void VariantCallFormat::VCF2genotypes(String &genofileName, String &subjectsFile){
	// read in all the subject ID and their columns
	VCFsubjects();
	// read in the selected subject ID
	DataFromFiles subjectIdData;
	subjectIdData.IDReadAsString(subjectsFile);
	// find out the columns to be output
	subjectSelected = subjectIdData.GetIndvsIDs();

	for (int i = 0; i < subjectSelected.size(); i++) {
		std::map<String,int>::iterator idx = subjectColumn.find(subjectSelected[i]);
		columnsSelected.push_back(idx->second);
	}
	
	// read in the genotype matrix
	FILE *vcfIn = fopen(m_vcfName, "r");
	FILE *genoOut = fopen(genofileName + ".012","w");
	FILE *genoPos = fopen(genofileName + ".012.pos", "w");
	FILE *genoIndv = fopen(genofileName + ".012.indv", "w");
	FILE *genoAF = fopen(genofileName + ".012.frq", "w");
	
	if (vcfIn == NULL) {
		printf("Error in opening file (%s)\n",m_vcfName.c_str());
		throw;
	}
	
	if (genoOut == NULL){
		printf("Error in opening file (%s)\n",genofileName.c_str());
		throw;
	}
	
	fileStep = 0;

	while (!feof(vcfIn)) {
		fileEntry.ReadLine(vcfIn);
		if (fileEntry.IsEmpty()) {
			printf("Error: File %s has an empty line (%d)\n",m_vcfName.c_str(),fileStep);
			throw;
		}
		if (fileEntry.Left(6) == "#CHROM") {
			// write the indv file -- file with subject IDs
			fileToken.AddTokens(fileEntry, "\t");
			for (int i = 0; i < columnsSelected.size(); i++) {
				fprintf(genoIndv, "%s\n", fileToken[columnsSelected[i]].c_str());
			}
			fclose(genoIndv); // can just use the subject ID files 
		}
		if (fileEntry.Left(1) != "#") { // genotype contents start from the lines without # 
			fileToken.AddTokens(fileEntry, "\t");
			// write the position file
			fprintf(genoPos, "%s\t%s\n", fileToken[0].c_str(), fileToken[1].c_str());
			// write the genotypes and initialize the genoSum counts
			genoSum = 0;
			nSample = 0;
			for (int i = 0; i < columnsSelected.size(); i++) {
				// get the genotype
				genoToken.AddTokens(fileToken[columnsSelected[i]], ":");
				//printf("genotype = %s\n",genoToken[0].c_str());
				alleleToken.AddTokens(genoToken[0], "/"); 
				
				// record the genotype -- check if it is misisng first
				if (!IsMissing()) {
					genoTypeNum = alleleToken[0].AsInteger() + alleleToken[1].AsInteger();	
					genoSum += genoTypeNum;
					nSample += 1;
				} else {
					genoTypeNum = -1;
				}
				
				// write out the genotype
				fprintf(genoOut, "%ld\t", genoTypeNum);
				
				// increment the file length
				fileStep = fileStep + 1;
				
				// clear the genoToken
				genoToken.Clear();
				alleleToken.Clear();
			}
			// write out the allele frequency (ID, ref, alt, af)
			double dSample = static_cast<double>(nSample);
			double dGenoSum = static_cast<double>(genoSum);
			fprintf(genoAF, "%s\t%s\t%s\t%1.5e\n", fileToken[idCol].c_str(), fileToken[refCol].c_str(), 
							fileToken[altCol].c_str(), 0.5*dGenoSum/dSample);
			fprintf(genoOut, "\n");
		}
		
		fileEntry.Clear();
		fileToken.Clear();
	}
		
	fclose(genoOut);
	fclose(genoPos);
	fclose(genoIndv);
	fclose(vcfIn);
	fclose(genoAF);
	
}

void VariantCallFormat::VCF2genotypesNonsynStop(String &genofileName){
	FILE *vcfIn = fopen(m_vcfName, "r");
	FILE *genoOut = fopen(genofileName + ".012","w");
	FILE *genoPos = fopen(genofileName + ".012.pos", "w");
	FILE *genoIndv = fopen(genofileName + ".012.indv", "w");
	FILE *genoAF = fopen(genofileName + ".012.frq", "w");
	
	if (vcfIn == NULL) {
		printf("Error in opening file (%s)\n",m_vcfName.c_str());
		throw;
	}
	
	if (genoOut == NULL){
		printf("Error in opening file (%s)\n",genofileName.c_str());
		throw;
	}
	
	fileStep = 0;
	while (!feof(vcfIn)) {
		fileEntry.ReadLine(vcfIn);
		if (fileEntry.IsEmpty()) {
			printf("Error: File %s has an empty line (%d)\n",m_vcfName.c_str(),fileStep);
			throw;
		}
		if (fileEntry.Left(6) == "#CHROM") {
			// write the indv file -- file with subject IDs
			fileToken.AddTokens(fileEntry, "\t");
			for (int i = 9; i < fileToken.Length(); i++) {
				fprintf(genoIndv, "%s\n", fileToken[i].c_str());
			}
			fclose(genoIndv);
		}
		
		// genotype contents start from the lines without # 
		if (fileEntry.Left(1) != "#") { 
			fileToken.AddTokens(fileEntry, "\t");
			// need to only look at line annotated at Nonsyn or Stop (loss/gain) 
			annotationInfo = fileToken[annotationCol].c_str();
			if (annotationInfo.find(searchNonsyn) != std::string::npos 
					|| annotationInfo.find(searchStop) != std::string::npos) {
				// write the position file
				fprintf(genoPos, "%s\t%s\n", fileToken[0].c_str(),fileToken[1].c_str());
				// write the genotypes and initialize the genoSum counts
				genoSum = 0;
				nSample = 0;
				for (int i = genoStartCol; i < fileToken.Length(); i++) {
					// get the genotype
					genoToken.AddTokens(fileToken[i], ":");
					//printf("genotype = %s\n",genoToken[0].c_str());
					alleleToken.AddTokens(genoToken[0], "/"); 
					
					// record the genotype -- check if it is misisng first
					if (!IsMissing()) {
						genoTypeNum = alleleToken[0].AsInteger() + alleleToken[1].AsInteger();	
						genoSum += genoTypeNum;
						nSample += 1;
					} else {
						genoTypeNum = -1;
					}
					
					// write out the genotype
					fprintf(genoOut, "%ld\t", genoTypeNum);
					
					// increment the file length
					fileStep = fileStep + 1;
					
					// clear the genoToken
					genoToken.Clear();
					alleleToken.Clear();
				}
				// write out the allele frequency (ID, ref, alt, af)
				double dSample = static_cast<double>(nSample);
				double dGenoSum = static_cast<double>(genoSum);
				fprintf(genoAF, "%s\t%s\t%s\t%1.5e\n", fileToken[idCol].c_str(), fileToken[refCol].c_str(), 
								fileToken[altCol].c_str(), 0.5*dGenoSum/dSample);
				fprintf(genoOut, "\n");
			}
		}
		
		fileEntry.Clear();
		fileToken.Clear();
	}
	// close the files that have been opened
	fclose(genoOut);
	fclose(genoPos);
	fclose(genoIndv);
	fclose(vcfIn);
	fclose(genoAF);	
}

void VariantCallFormat::VCF2genotypesNonsynStop(String &genofileName, String &subjectsFile){
	// read in all the subject ID and their columns
	VCFsubjects();
	// read in the selected subject ID
	DataFromFiles subjectIdData;
	subjectIdData.IDReadAsString(subjectsFile);
	// find out the columns to be output
	subjectSelected = subjectIdData.GetIndvsIDs();
	
	for (int i = 0; i < subjectSelected.size(); i++) {
		std::map<String,int>::iterator idx = subjectColumn.find(subjectSelected[i]);
		columnsSelected.push_back(idx->second);
	}
	
	// read in the genotype matrix
	FILE *vcfIn = fopen(m_vcfName, "r");
	FILE *genoOut = fopen(genofileName + ".012","w");
	FILE *genoPos = fopen(genofileName + ".012.pos", "w");
	FILE *genoIndv = fopen(genofileName + ".012.indv", "w");
	FILE *genoAF = fopen(genofileName + ".012.frq", "w");
	
	if (vcfIn == NULL) {
		printf("Error in opening file (%s)\n",m_vcfName.c_str());
		throw;
	}
	
	if (genoOut == NULL){
		printf("Error in opening file (%s)\n",genofileName.c_str());
		throw;
	}
	
	fileStep = 0;
	while (!feof(vcfIn)) {
		fileEntry.ReadLine(vcfIn);
		if (fileEntry.IsEmpty()) {
			printf("Error: File %s has an empty line (%d)\n",m_vcfName.c_str(),fileStep);
			throw;
		}
		if (fileEntry.Left(6) == "#CHROM") {
			// write the indv file -- file with subject IDs
			fileToken.AddTokens(fileEntry, "\t");
			for (int i = 0; i < columnsSelected.size(); i++) {
				fprintf(genoIndv, "%s\n", fileToken[columnsSelected[i]].c_str());
			}
			fclose(genoIndv); // can just use the subject ID files 
		}
		if (fileEntry.Left(1) != "#") { // genotype contents start from the lines without # 
			fileToken.AddTokens(fileEntry, "\t");
			// record the annotation Information string
			annotationInfo = fileToken[annotationCol].c_str();
			if (annotationInfo.find(searchNonsyn) != std::string::npos 
					|| annotationInfo.find(searchStop) != std::string::npos) {
				// write the position file
				fprintf(genoPos, "%s\t%s\n", fileToken[0].c_str(), fileToken[1].c_str());
				// write the genotypes and initialize the genoSum counts
				genoSum = 0;
				nSample = 0;
				for (int i = 0; i < columnsSelected.size(); i++) {
					// get the genotype
					genoToken.AddTokens(fileToken[columnsSelected[i]], ":");
					//printf("genotype = %s\n",genoToken[0].c_str());
					alleleToken.AddTokens(genoToken[0], "/"); 
					
					// record the genotype -- check if it is misisng first
					if (!IsMissing()) {
						genoTypeNum = alleleToken[0].AsInteger() + alleleToken[1].AsInteger();	
						genoSum += genoTypeNum;
						nSample += 1;
					} else {
						genoTypeNum = -1;
					}
					
					// write out the genotype
					fprintf(genoOut, "%ld\t", genoTypeNum);
										
					// clear the genoToken
					genoToken.Clear();
					alleleToken.Clear();
				}
				// increment the file length
				fileStep = fileStep + 1;
				
				// write out the allele frequency (ID, ref, alt, af)
				double dSample = static_cast<double>(nSample);
				double dGenoSum = static_cast<double>(genoSum);
				fprintf(genoAF, "%s\t%s\t%s\t%1.5e\n", fileToken[idCol].c_str(), fileToken[refCol].c_str(), 
								fileToken[altCol].c_str(), 0.5*dGenoSum/dSample);
				fprintf(genoOut, "\n");
			}
		}
		
		fileEntry.Clear();
		fileToken.Clear();
	}
	
	fclose(genoOut);
	fclose(genoPos);
	fclose(genoIndv);
	fclose(vcfIn);
	fclose(genoAF);	
}


void VariantCallFormat::VCF2TransformedGenoNonsynStop(String &genofileName, bool subsetSubject, String &subjectsFile, int posAll){
	// check the number of output variants
	if (posAll) {
		VCFpositions();
	} else {
		VCFpositionsNonsynStop();
	}
	
	FILE *genoIndv = fopen(genofileName + ".012.indv", "w");
	// columnsSelected would be either the size of subjectSelected or the total number of subject in the VCF file
	if (subsetSubject) {
		for (int i = 0; i < subjectSelected.size(); i++) {
			std::map<String,int>::iterator idx = subjectColumn.find(subjectSelected[i]);
			columnsSelected.push_back(idx->second);
			fprintf(genoIndv, "%s\n",(idx->first).c_str()); // it shouldn't be subjectSelected[i] b/c the subjectSelected might have more subject
		}
	} else {
		for	(std::map<String,int>::iterator i = subjectColumn.begin(); i != subjectColumn.end(); i++){
			columnsSelected.push_back(i->second); // note: since we've done this, need to ouptut the order of subject as in subjectColumn, and it should be the same order as the original vcf subject row header
			fprintf(genoIndv, "%s\n",(i->first).c_str());
		}
	} 
	
	// allocate a matrix of size numSubject by numPos
	Matrix tmpMatGenoStore(nSubjects,numPos);
	tmpMatGenoStore.Set(-1);

	// read in the genotype matrix
	FILE *vcfIn = fopen(m_vcfName, "r");
	FILE *genoOut = fopen(genofileName + ".012","w");
	FILE *genoPos = fopen(genofileName + ".012.pos", "w");
	FILE *genoAF = fopen(genofileName + ".012.frq", "w");

	// output allele frequency file with header: CHROM POS REF:Frequency ALT:Frequency
	fprintf(genoAF, "%s\t%s\t%s\t%s\t%s\n", "chrm", "pos","rsName", "ref","alt");
	
	if (vcfIn == NULL) {
		printf("Error in opening file (%s)\n",m_vcfName.c_str());
		throw;
	}
	
	if (genoOut == NULL){
		printf("Error in opening file (%s)\n",genofileName.c_str());
		throw;
	}
		
	fileStep = 0; // this fileStep is used to step into the tm geno matrix
	while (!feof(vcfIn)) {
		fileEntry.ReadLine(vcfIn);
		if (fileEntry.IsEmpty()) {
			printf("Error: File %s has an empty line (%d)\n",m_vcfName.c_str(),fileStep);
			throw;
		}
		if (fileEntry.Left(1) != "#") {
			fileToken.AddTokens(fileEntry, "\t");
			// posAll is to denote whether the position should be include in the output or not 
			// If there is a need to change the search criterion, I can modify this part
			if (posAll == 1) {
				posInclude = true;
			} else if (posAll == 0) {
				// record the annotation Information string
				annotationInfo = fileToken[annotationCol].c_str();
				posInclude = annotationInfo.find(searchNonsyn) != std::string::npos || annotationInfo.find(searchStop) != std::string::npos;
			} else {
				printf("Invalid posAll specification. It should be either 0 or 1.\n");
				throw;
			}

			if (posInclude) {
				// write the position file
				fprintf(genoPos, "%s\t%s\n", fileToken[0].c_str(), fileToken[1].c_str());
				// write the genotypes and initialize the genoSum counts
				genoSum = 0;
				nSample = 0;
				for (int i = 0; i < columnsSelected.size(); i++) {
					// get the genotype
					genoToken.AddTokens(fileToken[columnsSelected[i]], ":");
					//printf("genotype = %s\n",genoToken[0].c_str());
					alleleToken.AddTokens(genoToken[0], "/"); 
					
					// record the genotype -- check if it is misisng first
					if (!IsMissing()) {
						genoTypeNum = alleleToken[0].AsInteger() + alleleToken[1].AsInteger();	
						genoSum += genoTypeNum;
						nSample += 1;
					} else {
						genoTypeNum = -1;
					}
					
					// write out the genotype into the matrix
					tmpMatGenoStore[i][fileStep] = genoTypeNum;
					genoToken.Clear();
					alleleToken.Clear();
				}	
				fileStep = fileStep + 1 ;
				// write out the allele frequency (ID, ref, alt, af)
				double dSample = static_cast<double>(nSample);
				double dGenoSum = static_cast<double>(genoSum);
				double af = 0.5*dGenoSum/dSample;
				fprintf(genoAF, "%s\t%s\t%s\t%s:%1.5e\t%s:%1.5e\n", fileToken[0].c_str(), fileToken[1].c_str(), fileToken[2].c_str(),		
								fileToken[refCol].c_str(), 1-af, fileToken[altCol].c_str(), af);				
			}
		}
		fileEntry.Clear();
		fileToken.Clear();
	}
	fclose(genoPos);
	fclose(genoIndv);
	fclose(vcfIn);
	fclose(genoAF);	
	
	// now, output the 012 matrix into file -- we could skip this step and use the matrix in the RVtest
	// but for now, we decide to store the matrix first and re-read it.
	for (int i = 0; i < nSubjects; i ++) {
		for (int j = 0; j < numPos; j ++) {
			fprintf(genoOut, "%1.0f\t", tmpMatGenoStore[i][j]);
		}
		fprintf(genoOut, "\n");
	}
	fclose(genoOut);
}
