/*
 *  VariantCallFiles.h
 *  vcfReader
 *
 *  Created by Youna Hu on 1/17/11.
 *  Copyright 2011 U of Michigan. All rights reserved.
 *
 */

#include <vector>
#include <map>
#include "StringBasics.h"
#include "StringArray.h"
#include <string> 
#include "DataFromFiles.h"
#include "Error.h"

// if I need to generalize this piece of code, I shall let user input search string, 
// such as ;NONSY_:;STOP_ etc as parse it then search each individually.
// here for the purpose of extracting nonsyn SNPs, it is not necessary

class VariantCallFormat{ // this handles the vcf after Yanming's annotation
public:
	String m_vcfName;
	int fileStep;
	String fileEntry;
	std::string searchNonsyn,searchStop;
	std::string annotationInfo;
	StringArray fileToken;
	StringArray genoToken;
	StringArray alleleToken;
	long int genoTypeNum;
	std::map<String, int> subjectColumn;
	std::vector<String> subjectSelected;
	std::vector<int> columnsSelected;
	int genoColumn;
	long int genoSum;
	int nSample; // number of samples at each position without missing genotype
	bool sampleByPos;
	int genoStartCol,idCol,refCol,altCol,annotationCol;
	int nSubjects; // number of subjects -- note: This is different from nSamples
	int numPos; // number of samples
	bool posInclude;
	
	// update VCFsubjects and corresponding columns -- only need to read in the line starts with #CHROM
	void VCFsubjects();
	
	// update the number of positions (for writing into matrices)
	void VCFnumOutSubject(); // total number of subjects included in the VCF file
	void VCFnumOutSubject(String &subjectsFile); // total number of subjects specified by subjectsFile
	void VCFpositions(); // total number of variants (no screening)
	void VCFpositionsNonsynStop();  // total number of nonsyn + stop variants
	
	// Initializer
	void Init(String &vcfName);
	void Init(String &vcfName, String& subjectsFile);
	
	// constructor
	VariantCallFormat(String &vcfName);
	VariantCallFormat(String &vcfName, String &subjectsFile);
		
	// subtract genotype from vcf
	bool IsMissing();
	// output all genotypes
	void VCF2genotypes(String &genofileName);
	// output all genotypes for subjects specified by subjectsFile
	void VCF2genotypes(String &genofileName, String &subjectsFile); 
	
	// output all nonsynonymous and stop loss/gain annotated variants
	void VCF2genotypesNonsynStop(String &genofileName);
	// output all nonsyn + stop annotated variants for subjects specified by subjectsFile
	void VCF2genotypesNonsynStop(String &genofileName, String &subjectsFile);	
	
	// output the Nonsyn and stop genotypes in a transformed way (posAll: true (all), false( nonsyn + stop)
	void VCF2TransformedGenoNonsynStop(String &genofileName, bool subsetSubject, String &subjectsFile, int posAll);	
};